#include "expression.h"

// Allowed value ranges for the minute, hour, day and month field
const uint8_t min[4] = {0, 0, 1, 1};
const uint8_t max[4] = {59, 23, 31, 12};

// Convert a string a uint8_t value by parsing it using atoi and checking
// whether it's contained within the given range
#define SAFE_ATOI(v,s,min,max) \
	int _##v = atoi(s); \
	if ((_##v) == 0 && strcmp((s), "0") != 0) { \
		return CPEParseInvalidNumber; \
	} \
	if (v < (min) || v > (max)) { \
		return CPEParseOutOfRange; \
	} \
	v = (uint8_t) (_##v);

/**
 * Given a range expression, produce a bit field defining what numbers in the
 * min-max range the expression represents. Bit 0 (starting from the
 * right) corresponds to min, the bit max - min to max. All trailing bits
 * after this should be ignored. The given bitfield is modified in-place, so
 * multiple calls of this function can be performed on the same value to create
 * the effect of ORing their values.
 *
 * A range expression has one of the following forms:
 *
 * - *
 * - a
 * - a-b
 * - a/c
 * - a-b/c
 */
enum cron_parse_error ce_parse_range(uint64_t *out, char *s, uint8_t min, uint8_t max) {
	// The * expression means "every possible value"
	if (s[0] == '*') {
		// A '*' is only valid on its own
		if (s[1] != '\0') {
			return CPEParseInvalidExpression;
		}

		*out = ~0;

		return CPEParseOk;
	}

	size_t slash_index = 0;
	size_t dash_index = 0;
	size_t i = 0;

	// We first iterate over the string to determine whether it contains a slash
	// and/or a dash. We know the dash can only be valid if it appears before
	// the slash.
	while (s[i] != '\0' && slash_index == 0) {
		if (s[i] == '/') {
			slash_index = i;

			s[i] = '\0';
		} else if (s[i] == '-') {
			dash_index = i;

			s[i] = '\0';
		}

		i++;
	}

	// Parse the three possible numbers in the pattern
	uint8_t start = 0;
	uint8_t end = max;
	uint8_t interval = 1;

	SAFE_ATOI(start, s, min, max);

	if (dash_index > 0) {
		SAFE_ATOI(end, &s[dash_index + 1], min, max);
	}

	if (slash_index > 0) {
		SAFE_ATOI(interval, &s[slash_index + 1], 1, max - min);
	}

	if (dash_index == 0 && slash_index == 0) {
		*out |= ((uint64_t) 1) << (start - min);
	} else {
		while (start <= end) {
			*out |= ((uint64_t) 1) << (start - min);
			start += interval;
		}
	}

	return CPEParseOk;
}

/*
 * Given an expression part, produce a bitfield defining what numbers in the
 * min-max range the part represents. A part consists of one or more range
 * expressions, separated by commas.
 */
enum cron_parse_error ce_parse_part(uint64_t *out, char *s, uint8_t min, uint8_t max) {
	*out = 0;

	char *next;
	enum cron_parse_error res;

	while ((next = strchr(s, ',')) != NULL) {
		next[0] = '\0';
		res = ce_parse_range(out, s, min, max);

		if (res != CPEParseOk) {
			return res;
		}

		s = next + 1;
	}

	// Make sure to parse the final range as well
	return ce_parse_range(out, s, min, max);
}

/*
 * Return how many bits are set in the bitfield, better known as popcount. I
 * added my own implementation (taken from my algorithms course) as I don't want
 * to be dependent on GCC-specific extensions.
 */
uint8_t uint64_t_popcount(uint64_t n) {
	uint8_t c = 0;

	while (n != 0) {
		// This sets the least significant bit to zero (very cool)
		n &= n - 1;
		c++;
	}

	return c;
}

/*
 * Convert a bitfield into an array containing the numbers in the min-max range
 * it represents.
 */
uint8_t bf_to_nums(uint8_t **out, uint64_t bf, uint8_t min, uint8_t max) {
	// Each bit field only has `max - min + 1` meaningful bits. All other bits
	// should be ignored, and can be any value. By shifting the bit field back and
	// forth, we set these excessive bits to zero, ensuring popcount returns the
	// correct value.
	uint8_t excess_bits = 64 - (max - min + 1);
	bf = (bf << excess_bits) >> excess_bits;
	uint8_t size = uint64_t_popcount(bf);
	uint8_t *buf = malloc(size * sizeof(uint8_t));

	uint8_t i = 0, j = 0;

	while (j < size && i <= max - min) {
		if (((uint64_t)1 << i) & bf) {
			// Resize buffer if needed
			buf[j] = min + i;
			j++;
		}

		i++;
	}

	*out = buf;

	return size;
}

/*
 * Parse a cron expression string into a cron_expression struct.
 */
enum cron_parse_error ce_parse_expression(struct cron_expression *out, char *s) {
	// The parsing functions modify the input string in-place
	s = strdup(s);
	char *orig_s = s;

	uint8_t part_count = 0;

	char *next;
	enum cron_parse_error res = CPEParseOk;
	uint64_t bfs[4];

	// Skip leading spaces
	while (s[0] == ' ') {
		s++;
	}

	while (part_count < 4 && ((next = strchr(s, ' ')) != NULL)) {
		next[0] = '\0';
		res = ce_parse_part(&bfs[part_count], s, min[part_count], max[part_count]);

		if (res != CPEParseOk) {
			goto end;
		}

		size_t offset = 1;

		// Skip multiple spaces
		while (next[offset] == ' ') {
			offset++;
		}
		s = next + offset;

		part_count++;
	}

	// Parse final trailing part
	if (part_count < 4 && s[0] != '\0') {
		res = ce_parse_part(&bfs[part_count], s, min[part_count], max[part_count]);

		if (res != CPEParseOk) {
			goto end;
		}

		part_count++;
	}

	// At least two parts need to be provided
	if (part_count < 2) {
		res = CPEParseInvalidExpression;
		goto end;
	}

	// Ensure there's always 4 parts, as expressions can have between 2 and 4 parts
	while (part_count < 4) {
		// Expression is augmented with '*' expressions
		bfs[part_count] = ~0;
		part_count++;
	}

	out->minute_count = bf_to_nums(&out->minutes, bfs[0], min[0], max[0]);
	out->hour_count = bf_to_nums(&out->hours, bfs[1], min[1], max[1]);
	out->day_count = bf_to_nums(&out->days, bfs[2], min[2], max[2]);
	out->month_count = bf_to_nums(&out->months, bfs[3], min[3], max[3]);

end:
	// s is cloned
	free(orig_s);

	return res;
}
